# coding=utf-8

"""Extract content from DODFS and export to JSON.

Contains class ContentExtractor which have to public functions
avaiable to extract the DODF to JSON

Usage example::

    from dodfminer.extract.pure.core import ContentExtractor

    pdf_text = ContentExtractor.extract_text(file)
    ContentExtractor.extract_to_txt(folder)

"""

import os
import json
import unicodedata
import numpy as np

from pathlib import Path

import fitz

from dodfminer.extract.pure.utils.title_extractor import ExtractorTitleSubtitle
from dodfminer.extract.pure.utils.box_extractor import get_doc_text_boxes, sort_blocks

RESULTS_PATH = "results/"
RESULTS_PATH_JSON = "results/json"
RESULTS_PATH_TXT = "results/txt"


class ContentExtractor:
    """Extract content from DODFs and export to JSON.

    Extracts content from DODF files using as suport the title and subtitle
    databases—which runs using MuPDF—, and the Tesseract OCR library. All the
    content is exported to a JSON file, in which its keys are DODF titles or
    subtitles, and its values are the correspondent content.

    Note:
        This class is not constructable, it cannot generate objects.

    """

    @classmethod
    # pylint: disable=too-many-arguments
    def extract_text(cls, file, single=False, block=False, is_json=True, sep=' ', norm='NFKD'):
        """Extract block of text from file

        Args:
            file: The DODF to extract titles from.
            single: output content in a single file in the file directory.
            block: Extract the text as a list of text blocks.
            json: The list of text blocks are written as a json file.
            sep: The separator character between each block of text.
            norm: Type of normalization applied to the text.

        Note:
            To learn more about the each type of normalization used in the
            `unicode.normalization` method, `click here <https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize>`_.

        Returns:

            These are the outcomes for each parameter combination.

            When `block=True` and `single=True`:
                In case `json=True`, The method saves a JSON file containing the
                text blocks in the DODF file. However, is case `json=False`, the
                text from the whole PDF is saved as a string in a .txt file.

            When `block=True` and `single=False`:
                The method returns an array containing text blocks.

                Each array in the list have 5 values: the first four are the
                coordinates of the box from where the text was extracted
                (x0, y0, x1, y1), while the last is the text from the box.

                Example::

                    (127.77680206298828,
                    194.2507781982422,
                    684.0039672851562,
                    211.97523498535156,
                    "ANO XLVI EDICAO EXTRA No- 4 BRASILIA - DF")

            When `block=False` and `single=True`:
                The text from the whole PDF is saved in a .txt file as a
                normalized string.

            When `block=False` and `single=False`:
                The method returns a normalized string containing the
                text from the whole PDF.

        """
        drawboxes_text = ''
        list_of_boxes = []
        pymu_file = fitz.open(file)

        for page_boxes in get_doc_text_boxes(pymu_file):
            for text in page_boxes:
                if int(text[1]) != 55 and int(text[1]) != 881:
                    if block:
                        norm_text = cls._normalize_text(text[4], norm)
                        if is_json:
                            list_of_boxes.append((text[0], text[1], text[2],
                                                  text[3], norm_text))
                        else:
                            drawboxes_text += (norm_text + sep)
                    else:
                        drawboxes_text += (text[4] + sep)

        if block:
            if not single:
                return list_of_boxes
            if is_json:
                cls._save_single_file(file, 'json', json.dumps(list_of_boxes))
            else:
                return cls._save_single_file(file, 'txt', drawboxes_text)

        drawboxes_text = cls._normalize_text(drawboxes_text, norm)
        return drawboxes_text if not single else cls._save_single_file(file, 'txt', drawboxes_text)

    @classmethod
    def extract_structure(cls, file, single=False, norm='NFKD'):
        """Extract boxes of text with their respective titles.

        Args:
            file: The DODF file to extract titles from.
            single: Output content in a single file in the file directory.
            norm: `Type of normalization <https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize>`_ applied to the text.

        Returns:
            A dictionaty with the blocks organized by title.

            Example::

                {
                    "Title": [
                        [
                            x0,
                            y0,
                            x1,
                            y1,
                            "Text"
                        ]
                    ],
                    ...
                }

        """
        content_dict = {}

        try:
            title_base = cls._extract_titles(file).json.keys()
            # Aqui eh realmente necessario pegar um eception generica
            # pylint: disable=broad-except
        except Exception as excpt:
            cls._log(excpt)
            return None
        else:
            boxes = cls.extract_text(file, block=True, norm=norm)
            first_title = False
            is_title = False
            actual_title = ''
            section = None

            for box in boxes:
                text = box[4]
                is_title = True

                if text in ["SECAO I", "SECAO II", "SECAO III"]:
                    section = text
                    if section not in content_dict.keys():
                        content_dict.update({section: {}})
                else:
                    for title in title_base:
                        text = text.replace("\n", " ")
                        title = title.replace("\n", " ")
                        normalized_title = cls._normalize_text(title, norm)

                        if text == normalized_title:
                            first_title = True
                            actual_title = normalized_title
                            if title not in content_dict[section].keys():
                                content_dict[section].update({normalized_title: []})
                            break
                        else:
                            is_title = False

                if first_title and not is_title:
                    if int(box[1]) != 55 and int(box[1]) != 881:
                        content_dict[section][actual_title].append(box[:5])

            return content_dict if not single else cls._save_single_file(file, 'json', json.dumps(content_dict))

    @classmethod
    def extract_to_txt(cls, folder='./', norm='NFKD'):
        """Extract information from DODF to a .txt file.

        For each PDF file in data/DODFs, the method extracts information from the
        PDF and writes it to the .txt file.

        Args:
            folder: The folder containing the PDFs to be extracted.
            norm: `Type of normalization <https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize>`_ applied to the text.

        """
        pdfs_path_list = cls._get_pdfs_list(folder)
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH))
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH_TXT))
        txt_path_list = cls._get_txt_list(folder)

        for file in pdfs_path_list:
            pdf_name = os.path.splitext(os.path.basename(file))[0]
            if pdf_name not in txt_path_list:
                cls._log(pdf_name)
                text = cls.extract_text(file, norm=norm)
                t_path = cls._struct_subfolders(file, False, folder)
                with open(t_path, "w", encoding='utf-8') as file:
                    file.write(text)
            else:
                cls._log("TXT already exists")

    @classmethod
    def extract_to_json(cls, folder='./',
                        titles_with_boxes=False, norm='NFKD'):
        """Extract information from DODF to JSON.

        Args:
            folder: The folder containing the PDFs to be extracted.
            titles_with_boxes: If True, the method builds a dict containing a list of tuples (similar to `extract_structure`).
            Otherwise, the method structures a list of tuples (similar to `extract_text`).
            norm: `Type of normalization <https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize>`_ applied to the text.

        Returns:
            For each PDF file in data/DODFs, extract information from the
            PDF and output it to a JSON file.

        """
        # Get list of all downloaded pdfs
        pdfs_path_list = cls._get_pdfs_list(folder)
        # Get list of existing json to not repeat work
        json_path_list = cls._get_json_list(folder)

        cls._create_single_folder(os.path.join(folder, RESULTS_PATH))
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH_JSON))

        for file in pdfs_path_list:
            pdf_name = os.path.splitext(os.path.basename(file))[0]
            # We do not want the system to repeat itself doing the same work
            if pdf_name not in json_path_list:
                # low cost extractions
                if os.path.getsize(file) < 30000000:  # Remove in future.
                    # Remove images that might still there from previous exec
                    cls._log(pdf_name)
                    if titles_with_boxes:
                        content = cls.extract_structure(file, norm=norm)
                    else:
                        content = cls.extract_text(file, block=True, norm=norm)
                    j_path = cls._struct_subfolders(file, True, folder)
                    with open(j_path, "w", encoding="utf-8") as file:
                        json.dump(content, file,
                                ensure_ascii=False)
            else:
                cls._log("JSON already exists")

    @classmethod
    def _save_single_file(cls, file_path, file_type, content):
        with open(file_path.replace('pdf', file_type), 'w+', encoding='utf-8') as file:
            file.write(content)

    @classmethod
    def _normalize_text(cls, text, form='NFKD'):
        """This method is used for text nomalization.

        Args:
            text: The text to be normalized.
            form: `Type of normalization <https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize>`_ applied to the text.

        Returns:
            A string with the normalized text.

        """
        normalized = unicodedata.normalize(form, text).encode(
            'ascii', 'ignore').decode('utf8')
        return normalized

    @classmethod
    def _extract_titles(cls, file):
        """Extract titles and subtitles from the DODF.

        Args:
            file: The DODF to extract the titles.

        Returns:
            An object of type ExtractorTitleSubtitle, in which have the
            attributes:

            titles: get all titles from PDF.
            subtitle: get all subtitles from PDF.

        Raises:
            Exception: error in extracting titles from PDF.

        """
        try:
            title_database = ExtractorTitleSubtitle(file)
            cls._log(file)
        except Exception as exct:
            cls._log(f"Error in extracting files from {file}: {exct}")
            raise
        else:
            return title_database

    @classmethod
    def _get_pdfs_list(cls, folder):
        """Get DODFs list from the path.

        Args:
            folder: The folder containing the PDFs to be extracted.

        Returns:
            A list of DODFS' PDFs paths.

        """
        pdfs_path_list = []
        for dir_path, _, file_names in os.walk(os.path.expanduser(os.path.join(folder))):
            for file in file_names:
                if '.pdf' in file:
                    pdfs_path_list.append(os.path.join(dir_path, file))

        return pdfs_path_list

    @classmethod
    def _get_json_list(cls, folder):
        """Get list of exisiting JSONs from the path.

        Args:
            folder: The folder containing the PDFs to be extracted.

        Returns:
            A list of all exisiting JSONs.

        """
        aux = []
        for dir_path, _, file_names in os.walk(os.path.expanduser(os.path.join(folder, RESULTS_PATH_JSON))):
            for file in file_names:
                aux.append(os.path.join(dir_path, file))

        json_path_list = []
        for file in aux:
            json_path_list.append(os.path.splitext(os.path.basename(file))[0])

        return json_path_list

    @classmethod
    def _get_txt_list(cls, folder):
        """Get list of exisiting .txt files from the path.

        Args:
            folder: The folder containing the PDFs to be extracted.

        Returns:
            A list of all exisiting .txt files.

        """
        aux = []
        for dir_path, _, file_names in os.walk(os.path.expanduser(os.path.join(folder, RESULTS_PATH_TXT))):
            for file in file_names:
                aux.append(os.path.join(dir_path, file))

        txt_path_list = []
        for file in aux:
            txt_path_list.append(os.path.splitext(os.path.basename(file))[0])

        return txt_path_list

    @classmethod
    def _struct_subfolders(cls, path, json_f, folder):
        """Creates a directory for the JSON files.

        This method structures the folder tree for the allocation of
        files the code is curretly dealing with.

        Args:
            path: The path to the extracted file.
            json_f (boolean): If True, the file will extracted to a JSON. Otherwise, it will be extrated to a .txt.
            folder: The folder containing the PDFs to be extracted.

        Raises:
            FileExistsError: The folder being created is already there.

        Returns:
            The path created for the JSON to be saved.

        """
        type_f = '.json' if json_f else '.txt'
        res_path = RESULTS_PATH_JSON if json_f else RESULTS_PATH_TXT

        path = path.replace(folder, "")
        splited = path.split('/')
        basename = splited[-1].split('.')
        basename = basename[0] + type_f
        splited[-1] = basename
        final_path = '/'.join(splited[1:])
        final_path = os.path.join(folder, res_path, final_path)
        path = Path(os.path.join(folder, res_path, *splited[:-1]))
        try:
            path.mkdir(parents=True)
        except FileExistsError:
            pass

        return final_path

    @classmethod
    def _create_single_folder(cls, path):
        """Create a single folder given the directory path.

        This function might create a folder, observe if the folder already
        exists, or raise an error if the folder cannot be created.

        Args:
            path: The path to be created.

        Raises:
            OSError: Error creating the directory.

        """
        if os.path.exists(path):
            cls._log(os.path.basename(path) + " folder already exist")
        else:
            try:
                os.mkdir(path)
            except OSError as error:
                cls._log("Exception during the directory creation")
                cls._log(str(error))
            else:
                basename = os.path.basename(path)
                cls._log(basename + " directory successful created")

    @classmethod
    def _log(cls, msg):
        """Print message from within the ContentExtractor class.

        Args:
            msg: String with message that should be printed out.
        """
        print(f"[EXTRACTOR] {msg}")



import re
import os
from typing import List, Match
import joblib
import pandas as pd
import numpy as np
from dodfminer.extract.polished.acts.base import Atos


DODF = r"(DODF|[Dd]i.rio\s+[Oo]ficial\s+[Dd]o\s+[Dd]istrito\s+[Ff]ederal)"
_EDICAO_DODF = r"(\b(?i:suplement(o|ar)|extra|.ntegra))\b."
TIPO_EDICAO = r"\b(?P<tipo>(?i:extra(\sespecial)?|suplement(ar|o)))\b"
DODF_TIPO_EDICAO = DODF + r"(?P<tipo_edicao>.{0,50}?)" + _EDICAO_DODF

MONTHS = (
    r'(?i:janeiro|fevereiro|mar.o|abril|maio|junho|'
    r'julho|agosto|setembro|outubro|novembro|dezembro)'
)

FLEX_DATE = r"(?P<date>\d+\s+(?:de\s*)?" + MONTHS + \
    r"\s*(?:de\s*)?\d+|\d+[.]\d+[.]\d+|\d+[/]\d+[/]\d+)"

DODF_NUM = r"(?i:DODF|[Dd]i.rio [Oo]ficial [Dd]o [Dd]istrito [Ff]ederal)\s*[\w\W]{0,3}?(?i:n?(.mero|[.roº]{1,4})?[^\d]+?)(?P<num>\d+)"

DODF_DATE = DODF + r"[^\n\n]{0,50}?(?i:de\s?)?" + FLEX_DATE

SIAPE = r"(?i:siape)\s*(?i:n?.?)\s*[-\d.Xx/\s]"

MATRICULA = r"(?i:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"

MATRICULA_GENERICO = r"(?<![^\s])(?P<matricula>([-\d.XxZzYz/\s]{1,})[.-][\dXxYy][^\d])"

MATRICULA_ENTRE_VIRGULAS = r"(?<=[A-ZÀ-Ž]{3})\s*,\s+([-\d.XxZzYz/\s]{3,}?),"

PAGE = r"((?i:p\.|p.ginas?|p.?gs?\.?\b)(?P<page_nums>.{0,}?)(?=[,;:]|\n|\s[A-Z]|$))"

SERVIDOR_NOME_COMPLETO = r"(?i:servidora?\b.{0,40}?)(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{7,})"

NOME_COMPLETO = r"(?P<name>[.'A-ZÀ-Ž\s]{8,})"

LOWER_LETTER = r"[áàâäéèẽëíìîïóòôöúùûüça-z]"
UPPER_LETTER = r"[ÁÀÂÄÉÈẼËÍÌÎÏÓÒÔÖÚÙÛÜÇA-Z]"

PROCESSO_MATCH = r"(?i:processo):?[^\d]{0,50}?(?P<processo>\d[-0-9./\s]*\d(?!\d))"
TIPO_DOCUMENTO = r"(?i:portaria|ordem de servi.o|instru..o)"

EMPTY_MATCH = r'(?!x)x'

class SemEfeitoAposentadoria(Atos):
    '''
    Classe para atos que tornam aposentadoria sem efeito
    '''

    _special_acts = [
        'data_documento',
        'tipo_edicao',
        'numero_dodf',
        'pagina_dodf',
        'nome',
        'cargo_efetivo',
        'matricula',
    ]

    _BAD_MATCH_WORDS = [
        "AVERBAR",
        "NOMEAR",
        "CONCEDER",
        "EXONERAR",
        "DESAVERBAR",
        "APOSTILAR",
        "RETIFICAR",
    ]

    def _pre_process_text(self, text):
        # Make sure words splitted accross lines are joined together
        no_split_word = text.replace('-\n', '-')
        return no_split_word.replace('\n', ' ')

    def __init__(self, file, backend, debug=False, extra_search=True,
                 nlp=None, max_length=2000):
        self._max_length = max_length
        self._debug = debug
        self._extra_search = extra_search
        # self._processed_text = self._pre_process_text(open(file).read())
        self._raw_matches = []
        self._nlp = nlp
        super().__init__(file, backend)

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/sem_efeito_apo.pkl'
        return joblib.load(f_path)

    def _load_seg_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/seg_models/sem_efeito_apo.pkl'
        return joblib.load(f_path)

    def _act_name(self):
        return "Atos tornados sem efeito - aposentadoria"

    def _props_names(self):
        return list(self._prop_rules())

    def _rule_for_inst(self):
        return (
            r"TORNAR SEM EFEITO"
            r"([^\n]+\n){0,10}?[^\n]*?"
            r"(tempo\sde\sservi.o|aposentadoria|aposentou|([Dd][Ee][Ss])?[Aa][Vv][Ee][Rr][Bb][Aa]..[Oo]|(des)?averb(ar?|ou))"
            r"[\d\D]{0,500}?[.]\s+"
            r"(?=[A-Z]{3})"
        )

    def _prop_rules(self):
        return {
            'tipo_documento': TIPO_DOCUMENTO,
            'numero_documento': EMPTY_MATCH,
            'data_documento': EMPTY_MATCH,
            'numero_dodf': EMPTY_MATCH,
            'data_dodf': DODF_DATE,
            'pagina_dodf': EMPTY_MATCH,
            'nome': EMPTY_MATCH,
            'matricula': EMPTY_MATCH,
            'matricula_SIAPE': EMPTY_MATCH,
            'cargo_efetivo': EMPTY_MATCH,
            'classe': EMPTY_MATCH,
            'padrao': EMPTY_MATCH,
            'quadro': EMPTY_MATCH,
            'orgao': EMPTY_MATCH,
            'processo_SEI': PROCESSO_MATCH,
        }

    def _find_instances(self) -> List[Match]:
        """Returns list of re.Match objects found on `self._text_no_crosswords`.

        Return:
            a list with all re.Match objects resulted from searching for
        """
        head = "TORNAR SEM EFEITO"
        end = " CAFEBABE"   # so that lookeahead does not become a problem
        # lis = self._processed_text.split(head)
        lis = self._text.split(head)
        lis = [
            re.search(self._inst_rule, head + tex + end)
            # content before first `head` occurrence does not matter
            for tex in lis[1:]
        ]
        lis = [i for i in lis if i]

        true_positive = []
        for raw_match in lis:
            flag = True
            for bad in self._BAD_MATCH_WORDS:
                s = raw_match.group()
                if len(s) > self._max_length or bad in s:
                    flag = False
                    break
            if flag:
                true_positive.append(raw_match)

        self._raw_matches = true_positive
        if self._debug:
            print("DEBUG:", len(lis), 'generic matches')
            print("DEBUG:", len(self._raw_matches), 'true matches')
        return [i.group() for i in self._raw_matches]

    def _get_special_acts(self, lis_dict):
        for i, match in enumerate(self._raw_matches):
            act = match.group()
            curr_dict = lis_dict[i]
            data_dodf = curr_dict['data_dodf']
            if data_dodf == 'nan' or not data_dodf:
                data_dodf = re.search(
                    FLEX_DATE, self._raw_acts[i]
                )
                curr_dict['data_dodf'] = data_dodf


            numero_dodf = data_dodf and re.search(DODF_NUM, data_dodf.group())
            data_documento = data_dodf and \
                re.search(
                    FLEX_DATE, act[:data_dodf.start()] + act[data_dodf.end():])
            pagina_dodf = data_dodf and re.search(
                PAGE, act[data_dodf.end():][:50])

            nome = re.search(SERVIDOR_NOME_COMPLETO, act)
            if self._debug:
                print('NOME ENCONTRADO:', nome)
            if not nome:
                #  If it fails then a more generic regex is searched for
                dodf_mt = re.search(DODF, act)
                dodf_end = 0 if not dodf_mt else dodf_mt.end()
                nome = re.search(NOME_COMPLETO, act[dodf_end:])
                if self._debug:
                    print("NOME ENCONTRADO 2:", nome)
                del dodf_mt, dodf_end
                if not nome and self._nlp:
                    # Appeal to spacy
                    all_cands = re.findall(NOME_COMPLETO, act)
                    cand_text = 'SEM-SERVIDOR'
                    for cand in self._nlp(', '.join([c.strip().title() for c in all_cands])).ents:
                        cand_text = cand.text

                        if cand.label_ == 'PER':
                            break
                    nome = re.search(cand_text.upper(), act)
                    del all_cands, cand_text, cand
            end_employee = nome.end() if nome else 0
            matricula = re.search(MATRICULA, act[end_employee:]) or \
                re.search(MATRICULA_GENERICO, act[end_employee:]) or \
                re.search(MATRICULA_ENTRE_VIRGULAS, act[end_employee:])

            del end_employee
            if not matricula or not nome:
                cargo_efetivo = None
            else:
                nome_start = act[nome.start():].find(
                    nome.group()) + nome.start()
                matricula_start = act[matricula.start():].find(
                    matricula.group()) + matricula.start()

                # NOTE: -1 is important in case `matricula` end with `,`
                if 0 <= (matricula_start - (nome_start + len(nome.group()))) <= 5:
                    # cargo_efetivo does not fit between 'nome' and 'matricula'
                    cargo_efetivo = re.search(
                        r",(?P<cargo_efetivo>[^,]+)", act[matricula_start + len(matricula.group())-1:])

                else:
                    # cargo_efetivo right after employee's name

                    cargo_efetivo = re.search(
                        r",(?P<cargo_efetivo>[^,]+)", act[nome_start + len(nome.group())-1:])
                del matricula_start, nome_start
            edicao = re.search(DODF_TIPO_EDICAO, act)
            tipo_edicao = re.search(TIPO_EDICAO, act[edicao.start()-1:edicao.end()+1])\
                if edicao else re.search("normal", "normal")
            curr_dict['data_documento'] = data_documento
            curr_dict['numero_dodf'] = numero_dodf
            curr_dict['pagina_dodf'] = pagina_dodf
            curr_dict['nome'] = nome
            curr_dict['matricula'] = matricula
            curr_dict['cargo_efetivo'] = cargo_efetivo
            curr_dict['tipo_edicao'] = tipo_edicao

    def _find_prop_value(self, rule, act):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        match = re.search(rule, act, flags=self._flags)
        return match,

    def _group_solver(self, match):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        if not match or isinstance(match, str):
            return np.nan
        elif match.groupdict():
            key = list(match.groupdict())[0]
            return match.group(key)
        else:
            return match.group()

    def _extract_props(self):
        acts = []
        for raw in self._raw_acts:
            act = self._regex_props(raw)
            acts.append(act)
        if self._extra_search:
            self._get_special_acts(acts)
        return acts

    def _regex_instances(self) -> List[Match]:
        found = self._find_instances()
        self._acts_str = found.copy()
        return found

    def _build_dataframe(self):
        _ = re.search(self._name, self._name)
        for dic in self._acts:
            dic["tipo_ato"] = _
        data = [{k: self._group_solver(v) for k, v in act.items()}
                for act in self._acts]
        return pd.DataFrame(data)



"""Functions to extract boxes from text."""

import fitz
import re
from functools import cmp_to_key, reduce
import operator

SECTION_TITLES = ["SEÇÃO I", "SEÇÃO II", "SEÇÃO III"]


def _extract_page_lines_content(page):
    """Extracts page lines.

    Args:
        page: fitz.fitz.Page object to have its bold content extracted.

    Returns:
        List[tuple(float, float, float, float, str)]
        A list containing lines content at the page, along with
        its bounding boxes.

    """
    lis = []
    for block in page.getTextPage().extractDICT()['blocks']:
        for line in block['lines']:
            for span in line['spans']:
                del span['color']
                del span['flags']
                lis.append((*span['bbox'], span['text']))
    return lis


def get_doc_text_boxes(doc: fitz.Document):
    """Returns list of list of extracted text blocks.

    Args:
        doc: an opened fitz document.

    Returns:
        List[List[tuple(float, float, float, float, str, int, int)]]

    """

    text_blocks = [sort_blocks(page.getTextBlocks()) for page in doc]
    return text_blocks


def sort_blocks(page_blocks):
    """Sort blocks by their vertical and horizontal position.

        Args:
            page_blocks: a list of blocks within a page.

        Returns:
            List[tuple(float, float, float, float, str, int, int)]
    """
    return sorted(page_blocks, key=cmp_to_key(compare_blocks))


def compare_blocks(block1, block2):
    """Implements a comparison heuristic between blocks.
       Blocks that are in the uppermost and leftmost positions 
       should be inserted before the other block in comparison.

    Args:
        block1: a block tuple to be compared.
        block2: a block tuple to be compared to.

    Returns:
        Int
    """
    b1_x0, b1_y0, b1_x1, b1_y1, *_ = block1
    b2_x0, b2_y0, b2_x1, b2_y1, *_ = block2

    b1_y = max([b1_y0, b1_y1])
    b2_y = max([b2_y0, b2_y1])

    if (b1_x0 >= 49 and b1_x1 <= 405 and b2_x0 >= 55 and b2_x1 <= 405) or \
       (b1_x0 >= 417 and b1_x1 <= 766 and b2_x0 >= 417 and b2_x1 <= 766) or \
       (b1_x1-b1_x0 > 350) or \
       (b2_x1-b2_x0 > 350) or \
       (b1_y < 80) or \
       (b2_y < 80):
        return b1_y-b2_y
    else:
        return b1_x0-b2_x0

def draw_doc_text_boxes(doc: fitz.Document, doc_boxes, save_path=None):
    """Draw extracted text blocks rectangles.
       In result, a pdf file with rectangles shapes added, representing the extracted blocks,
       is saved.

    Args:
        doc: an opened fitz document
        doc_boxes: the list of blocks on a document, separated by pages
        save_path: a custom path for saving the result pdf 

    Returns:
        None
    """
    color = fitz.utils.getColor("greenyellow")

    for page in doc:
        for page_box in doc_boxes[page.number]:
            x0, y0, x1, y1, *_ = page_box
            rect = fitz.Rect(x0, y0, x1, y1)

            page.drawRect(rect, color, width=2)

    doc_path = '/'.join(doc.name.split('/')[0:-1])
    doc_name = doc.name.split('/')[-1]

    if save_path != None:
        doc.save(f"{save_path}/BOXES_{doc_name}")
    else:
        doc.save(f"{doc_path}{'/' if len(doc_path) else ''}BOXES_{doc_name}")


def get_doc_text_lines(doc: fitz.Document):
    """Returns list of list of extracted text lines.

    Args:
        doc: an opened fitz document.

    Returns:
        List[List[tuple(float, float, float, str)]]

    """

    return [_extract_page_lines_content(page) for page in doc]


def _get_doc_img(doc: fitz.Document):
    """Returns list of list of image items.

    Note:
        This function is not intented to be used by final users,
        but internally. Image `items` are described at:

        https://pymupdf.readthedocs.io/en/latest/page/#Page.getImageBbox

    Args:
        doc: an opened fitz document

    Returns:
        List[List[tuple(int, int, int, int, str, str, str, str, int)]]
        (xref, smask, width, height, bpc, colorspace,
        alt. colorspace, filter, invoker)
    """

    return [page.getImageList(full=True) for page in doc]


def get_doc_img_boxes(doc: fitz.Document):
    """Returns list of list of bouding boxes of extracted images.

    Args:
        doc: an opened fitz document

    Returns:
        List[List[Rect(float, float, float, float)]]. Each Rect represents
            an image bounding box.

    """

    img_ll = _get_doc_img(doc)
    for idx, img_lis in enumerate(img_ll):
        page = doc[idx]
        img_list_altered = []

        for img in img_lis:
            img = img[:9] + (0,)
            img_list_altered.append(img)

        img_ll[idx] = [page.getImageBbox(img) for img in img_list_altered]

    return img_ll



"""NER backend for act and propriety extraction.

This module contains the ActNER class, which have all that is necessary to
extract an act and, its proprieties, using a trained ner model.

"""

import os
import nltk
import numpy as np
import joblib
import re

# pylint: disable=too-few-public-methods


class ActNER:
    """Act NER Class.

    This class encapsulate all functions, and attributes related
    to the process of NER extraction.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _model: The trained NER model for the act

    """

    def __init__(self):
        # self._backend = 'regex'
        nltk.download('punkt', quiet=True)
        super().__init__()

        # pylint: disable=assignment-from-no-return
        self._model = self._load_model()

    def _load_model(self):
        """Load Model from models/folder.

        Note:
            This function needs to be overwriten in
            the child class. If this function is not
            overwrite the backend will fall back to regex.

        """
        # pylint: disable=access-member-before-definition
        if self._backend == 'ner':
            print(f"Act {self._name} does not have an entity extraction model: FALLING BACK TO REGEX")
            self._backend = 'regex'
        else:
            self._backend = 'regex'

    def _prediction(self, act):
        """Predict classes for a single act.

        Args:
            act (string): Full act

        Returns:
            A dictionary with the proprieties and its
            predicted value.
        """
        act = self._preprocess(act)
        feats = self._get_features(self._split_sentence(act))
        pred = self._model.predict_single(feats)
        return self._predictions_dict(act, pred)

    @classmethod
    def _preprocess(cls, text):
        """Preprocess text for CRF model."""
        text = text.replace('\n', ' ').strip()
        text = re.sub(' +', ' ', text)
        text = re.sub(r'([a-zA-Z0-9])- ', r'\1', text)
        return text

    def _limits(self, sentence):
        """Find the limits of words in the sentence.
        
        Args:
            sentence (str): target sentence.
            
        Returns:
            List of the positions in which each word in sentence starts.
        """
        letters = [chr(c) for c in range(ord('a'), ord('z') + 1)]
        numbers = [chr(c) for c in range(ord('0'), ord('9') + 1)]
        symbols = ['(', ',', '.', '/', '-']
        all = letters + numbers + symbols + [' ']

        lim = []
        if sentence[0] != ' ':
            lim.append(0)
            
        for i in range(1, len(sentence)):
            current = sentence[i].lower()
            previous = sentence[i-1].lower()
            
            if current in letters and previous not in letters:
                lim.append(i)
            elif current in numbers and previous not in numbers:
                lim.append(i)
            elif current in symbols:
                lim.append(i)
            elif current not in all and previous in letters:
                lim.append(i)
        return lim


    def _split_sentence(self, sentence):
        """Split a sentence into words.
        
        Args:
            sentence (str): Sentence to be split.
            
        Returns:
            List of words in the sentence.
        """
        lim = self._limits(sentence)
        lim.append(len(sentence))
        
        words = []
        for i in range(1, len(lim)):
            words.append(sentence[lim[i-1]:lim[i]].strip())
        return words

    def _get_base_feat(self, word):
        """Get the base features of a word, for the CRF model.
        
        Args:
            word (str): Word to be processed.
            
        Returns:
            Dictionary with the base features of the word.
        """
        d = {
            'word': word.lower(),
            'is_title': word.istitle(),
            'is_upper': word.isupper(),
            'num_digits': str(sum(c.isdigit() for c in word)),
        }
        return d

    def _add_base_feat(self, features, sentence, index, prefix):
        """Updates a dictionary of features with the features of a word.
        
        Args:
            features (dict): Dictionary with the features already processed.
            sentence (list): List of words in the sentence.
            index (int): Index of the current word in the sentence. 
            prefix (str): Prefix to be added to the name of the features of the current word.

        """
        if index >= 0 and index < len(sentence):
            word_feat = self._get_base_feat(sentence[index])
            for feat in word_feat.keys():
                features[prefix + feat] = word_feat[feat]

    def _get_features(self, sentence):
        """Get the features of a sentence, for the CRF model.
        
        Args:
            sentence (list): List of words in the sentence.
            
        Returns:
            List of dictionaries with the features of each word.
        """
        sent_features = []
        
        for i in range(len(sentence)):
            word = sentence[i]

            word_feat = {
                'bias': 1.0,
                'text_position': i/len(sentence),
            }
            
            self._add_base_feat(word_feat, sentence, i-4, '-4:')
            self._add_base_feat(word_feat, sentence, i-3, '-3:')
            self._add_base_feat(word_feat, sentence, i-2, '-2:')
            self._add_base_feat(word_feat, sentence, i-1, '-1:')
            
            self._add_base_feat(word_feat, sentence, i, '')
            
            self._add_base_feat(word_feat, sentence, i+1, '+1:')
            self._add_base_feat(word_feat, sentence, i+2, '+2:')
            self._add_base_feat(word_feat, sentence, i+3, '+3:')
            self._add_base_feat(word_feat, sentence, i+4, '+4:')
                
            sent_features.append(word_feat)

        return sent_features

    def _predictions_dict(self, sentence, prediction):
        """Create dictionary of proprieties.

        Create dictionary of tags to save predicted entities.

        Args:
            sentence (list): List of words and tokens in the act.
            prediction ([type]): The correspondent predicitons for each
                                 word in the sentence.

        Returns:
            A dictionary of the proprieties found.

        """
        
        dict_ato = {}
        for c in self._model.classes_:
            if c == 'O':
                continue
            dict_ato[c[2:]] = []
            
        limits = self._limits(sentence)
        
        limits.append(len(sentence))
        prediction.append('O')

        current = ''
        entity_start = -1
        for i in range(len(prediction)):
            if current != '' and prediction[i] != 'I-' + current:
                entity_end = limits[i]
                dict_ato[current].append(sentence[entity_start:entity_end].strip())
                entity_start = -1
                current = ''

            if prediction[i][0] == 'B' or (prediction[i][0] == 'I' and current == ''):
                current = prediction[i][2:]
                entity_start = limits[i]

        for i in dict_ato.keys():
            if len(dict_ato[i]) == 0:
                dict_ato[i] = np.nan
            elif len(dict_ato[i]) == 1:
                dict_ato[i] = dict_ato[i][0]

        return dict_ato



# pylint: disable=protected-access

import os
from unittest.mock import patch
import joblib
import pytest
from dodfminer.extract.polished.backend.seg import ActSeg

# act_seg with regex segmentation

@pytest.fixture(name='act_seg_regex')
@patch.object(ActSeg, '_backend', create = True)
def fixture_act_seg_regex(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'regex'

    act._flags = 0
    act._rules = {
        "numeros": r"([0-9]+)",
        "capitalizado": r"([A-Z][a-z]+)",
        "nao_encontra": "khalil"
    }
    act._inst_rule = r"(MENSAGEM:)([^.]*)"
    act._acts_str = []

    act.__init__()
    return act

def test_act_seg_regex(act_seg_regex):
    assert act_seg_regex._backend == 'regex'
    assert act_seg_regex._seg_function == act_seg_regex._regex_instances

# act_seg with ner segmentation

@pytest.fixture(name='act_seg_ner')
@patch.object(ActSeg, '_backend', create = True)
@patch.object(ActSeg, '_load_seg_model',
              return_value=joblib.load(""+os.path.dirname(__file__)+"/support/test_seg_model.pkl"))
def fixture_act_seg_ner(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'ner'
    act.__init__()
    return act

def test_act_seg_ner(act_seg_ner):
    assert act_seg_ner._backend == 'ner'
    assert act_seg_ner._seg_function == act_seg_ner._crf_instances

# act_seg with ner background but no model, reverting to regex segmentation

@pytest.fixture(name='act_seg_reverting_to_regex')
@patch.object(ActSeg, '_backend', create = True)
def fixture_act_seg_reverting_to_regex(*_):
    act = ActSeg()
    act._name = 'test'
    act._backend = 'ner'
    act.__init__()
    return act

def test_act_seg_reverting_to_regex(act_seg_reverting_to_regex):
    assert act_seg_reverting_to_regex._backend == 'ner'
    assert act_seg_reverting_to_regex._seg_function == act_seg_reverting_to_regex._regex_instances

#

def test_act_seg_preprocess(act_seg_ner):
    sentence = "teste com\npalavras."
    assert act_seg_ner._preprocess(sentence) == 'teste com palavras.'

def test_act_seg_limits(act_seg_ner):
    sentence = "teste com palavras.123"
    assert act_seg_ner._limits(sentence) == [0, 6, 10, 18, 19]

def test_act_seg_regex_instances(act_seg_regex):
    act_seg_regex._text =  "fghngfnfgnfgnfgnrr MENSAGEM: O Renato testa esse codigo 1232 vezes por dia. qeqcnoecqucqpwxqrthrweqqeprto MENSAGEM: " + \
        "O Joao testa esse codigo 2 vezes por dia. wsfhsn wenrwermrkne MENSAGEM: O Lucas testa esse codigo 0 vezes por dia. efwefwefwfwefwetrynyujuju"
    act_seg_regex.acts_str = []
    act = act_seg_regex._seg_function()

    assert act == [' O Renato testa esse codigo 1232 vezes por dia',
                   ' O Joao testa esse codigo 2 vezes por dia', ' O Lucas testa esse codigo 0 vezes por dia']

def test_act_seg_crf_instances(act_seg_ner):
    act_seg_ner._text = ": CONCEDER, aposentadoria voluntaria integral, ao servidor ELIAS SANTOS MONTEIRO, matrícula nº 24.679-4, no cargo de " + \
        "Analista em Políticas Públicas e Gestão Governamental, Classe Especial, Padrão V, do Quadro de Pessoal do Distrito Federal, nos termos " + \
        "do artigo 3º, incisos I, II e III, e Parágrafo único da Emenda Constitucional nº 47 de 05/07/2005, combinado com o artigo 44 da Lei " + \
        "Complementar nº 769, de 30/06/2008, e com a vantagem pessoal prevista no artigo 5º da Lei nº 4.584, de 08/07/2011. Lotação: " + \
        "Administração Regional de Brazlândia. Processo SEI nº 00133-00002749/2020-19. CONCEDER, aposentadoria voluntaria integral, a servidora " + \
        "MIRIAM RODRIGUES DA SILVA, matricula nº 31.325-4, no cargo de Analista em Politicas Públicas e Gestao Governamental, Classe Especial, " + \
        "Padrão V, do Quadro de Pessoal do Distrito Federal, nos termos do artigo 3º, incisos I, II e III, e Parágrafo único da Emenda " + \
        "Constitucional nº 47 de 05/07/2005, combinado com o artigo 44 da Lei Complementar nº 769, de 30/06/2008, e com a vantagem pessoal " + \
        "prevista no artigo 5º da Lei nº 4.584, de 08/07/2011. Lotação: Administração Regional de Brazlândia. Processo SEI nº " + \
        "00133-00002735/2020-97."
    act_seg_ner._acts_str = []
    act = act_seg_ner._seg_function()

    assert len(act) == 2

def test_act_seg_split_sentences(act_seg_ner):
    text = "CONCEDER, aposentadoria voluntária integral, ao servidor ELIAS SANTOS MONTEIRO, matrícula nº 24.679-4, no cargo de " + \
        "Analista em Políticas Públicas e Gestão Governamental, nos termos..."
    preprocessed = ['CONCEDER', ',', 'aposentadoria', 'volunt', 'á', 'ria', 'integral', ',', 'ao', 'servidor', 'ELIAS', 'SANTOS', 'MONTEIRO', ',',
        'matr', 'í', 'cula', 'n', 'º', '24', '.', '679', '-', '4', ',', 'no', 'cargo', 'de', 'Analista', 'em', 'Pol', 'í', 'ticas', 'P', 'ú',
        'blicas', 'e', 'Gest', 'ã', 'o', 'Governamental', ',', 'nos', 'termos', '.', '.', '.']
    assert act_seg_ner._split_sentence(text) == preprocessed

def test_act_seg_get_base_feat(act_seg_ner):
    word = "NER"
    assert act_seg_ner._get_base_feat(word) == {
        'word': 'ner',
        'is_title': False,
        'is_upper': True,
        'num_digits': '0',
    }

def test_act_seg_add_base_feat(act_seg_ner):
    features = {'': 0}
    sentence = ["Teste001", "."]
    index = 1
    prefix = '+1:'

    feat_target = {'': 0, '+1:word': '.', '+1:is_title': False, '+1:is_upper': False, '+1:num_digits': '0'}

    act_seg_ner._add_base_feat(features, sentence, index, prefix)

    assert features == feat_target

def test_act_seg_get_features(act_seg_ner):
    sentence = ["Art", ".", "5", ",", "ATOS"]
    features_1 = [
        {
            'bias': 1.0,
            'text_position': 0.0,
            'word': 'art',
            'is_title': True,
            'is_upper': False,
            'num_digits': '0',
            '+1:word': '.',
            '+1:is_title': False,
            '+1:is_upper': False,
            '+1:num_digits': '0',
            '+2:word': '5',
            '+2:is_title': False,
            '+2:is_upper': False,
            '+2:num_digits': '1',
            '+3:word': ',',
            '+3:is_title': False,
            '+3:is_upper': False,
            '+3:num_digits': '0',
            '+4:word': 'atos',
            '+4:is_title': False,
            '+4:is_upper': True,
            '+4:num_digits': '0'
        },
        {
            'bias': 1.0,
            'text_position': 0.2,
            '-1:word': 'art',
            '-1:is_title': True,
            '-1:is_upper': False,
            '-1:num_digits': '0',
            'word': '.',
            'is_title': False,
            'is_upper': False,
            'num_digits': '0',
            '+1:word': '5',
            '+1:is_title': False,
            '+1:is_upper': False,
            '+1:num_digits': '1',
            '+2:word': ',',
            '+2:is_title': False,
            '+2:is_upper': False,
            '+2:num_digits': '0',
            '+3:word': 'atos',
            '+3:is_title': False,
            '+3:is_upper': True,
            '+3:num_digits': '0'
        },
        {
            'bias': 1.0,
            'text_position': 0.4,
            '-2:word': 'art',
            '-2:is_title': True,
            '-2:is_upper': False,
            '-2:num_digits': '0',
            '-1:word': '.',
            '-1:is_title': False,
            '-1:is_upper': False,
            '-1:num_digits': '0',
            'word': '5',
            'is_title': False,
            'is_upper': False,
            'num_digits': '1',
            '+1:word': ',',
            '+1:is_title': False,
            '+1:is_upper': False,
            '+1:num_digits': '0',
            '+2:word': 'atos',
            '+2:is_title': False,
            '+2:is_upper': True,
            '+2:num_digits': '0'
        },
        {
            'bias': 1.0,
            'text_position': 0.6,
            '-3:word': 'art',
            '-3:is_title': True,
            '-3:is_upper': False,
            '-3:num_digits': '0',
            '-2:word': '.',
            '-2:is_title': False,
            '-2:is_upper': False,
            '-2:num_digits': '0',
            '-1:word': '5',
            '-1:is_title': False,
            '-1:is_upper': False,
            '-1:num_digits': '1',
            'word': ',',
            'is_title': False,
            'is_upper': False,
            'num_digits': '0',
            '+1:word': 'atos',
            '+1:is_title': False,
            '+1:is_upper': True,
            '+1:num_digits': '0'
        },
        {
            'bias': 1.0,
            'text_position': 0.8,
            '-4:word': 'art',
            '-4:is_title': True,
            '-4:is_upper': False,
            '-4:num_digits': '0',
            '-3:word': '.',
            '-3:is_title': False,
            '-3:is_upper': False,
            '-3:num_digits': '0',
            '-2:word': '5',
            '-2:is_title': False,
            '-2:is_upper': False,
            '-2:num_digits': '1',
            '-1:word': ',',
            '-1:is_title': False,
            '-1:is_upper': False,
            '-1:num_digits': '0',
            'word': 'atos',
            'is_title': False,
            'is_upper': True,
            'num_digits': '0'
        }
    ]
    assert features_1 == act_seg_ner._get_features(sentence)


def test_act_seg_extract_acts(act_seg_ner):
    text = "OOOO AAAA BBBB CCCC DDDD EEEE FFFF XXXX YYYY ZZZZ"
    pred = "O B-Ato I-Ato E-ato O O B-Ato I-Ato E-Ato O".split()
    alvo =["AAAA BBBB CCCC",            "FFFF XXXX YYYY"]

    assert act_seg_ner._extract_acts(text, pred) == alvo



"""Base class for act segmentation.

This module contains the ActSeg class, which have all that is necessary to
extract acts from a block of text.
"""

import re
from nltk import word_tokenize


class ActSeg:
    """Base class for act segmentation.

    This class encapsulate all functions, and attributes related
    to the process of segmentation of an act.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _seg_function: Function for segmentation.

    """

    def __init__(self):
        self._seg_function = self._load_seg_function()

    def _load_seg_function(self):
        """Load segmentation function into the _seg_function variable.

        If the segmentation backend is ner and a NER model is available, loads the function _crf_instances.
        Otherwise, loads the function _regex_instances.

        """
        if self._backend == 'ner':
            self._seg_model = self._load_seg_model()
            if self._seg_model is not None:
                return self._crf_instances
            else:
                print(f"Act {self._name} does not have a segmentation model: FALLING BACK TO REGEX")
                return self._regex_instances
        else:
            self._backend = 'regex'
            return self._regex_instances

    def _load_seg_model(self):
        """Load Segmentation Model from models/folder.

        Note:
            This function needs to be overwriten in
            the child class. If this function is not
            overwrite the segmentation backend will fall back to regex.

        """
        return None

    def _regex_instances(self):
        """Search for all instances of the act using the defined rule.

        Returns:
            List of all act instances in the text.
        """

        # pylint: disable=no-member
        found = re.findall(self._inst_rule, self._text, flags=self._flags)
        results = []
        for instance in found:
            head, body, *_ = instance
            # pylint: disable=no-member
            self._acts_str.append(head+body)
            results.append(body)

        return results

    def _crf_instances(self):
        """Search for all instances of the act using a CRF model.

        Returns:
            List of all act instances in the text.
        """
        text = self._preprocess(self._text)
        feats = self._get_features(self._split_sentence(text))
        pred = self._seg_model.predict_single(feats)
        acts = self._extract_acts(text, pred)
        self._acts_str += acts
        return acts

    def _preprocess(self, text):
        """Preprocess text for CRF model."""
        return text.replace('\n', ' ').strip()

    def _limits(self, sentence):
        """Find the limits of words in the sentence.
        
        Args:
            sentence (str): target sentence.
            
        Returns:
            List of the positions in which each word in sentence starts.
        """

        letters = [chr(c) for c in range(ord('a'), ord('z') + 1)]
        numbers = [chr(c) for c in range(ord('0'), ord('9') + 1)]
        symbols = ['(', ',', '.', '/', '-']
        all = letters + numbers + symbols + [' ']

        lim = []
        if sentence[0] != ' ':
            lim.append(0)
            
        for i in range(1, len(sentence)):
            current = sentence[i].lower()
            previous = sentence[i-1].lower()
            
            if current in letters and previous not in letters:
                lim.append(i)
            elif current in numbers and previous not in numbers:
                lim.append(i)
            elif current in symbols:
                lim.append(i)
            elif current not in all and previous in letters:
                lim.append(i)
        return lim

    def _split_sentence(self, sentence):
        """Split a sentence into words.
        
        Args:
            sentence (str): Sentence to be split.
            
        Returns:
            List of words in the sentence.
        """
        lim = self._limits(sentence)
        lim.append(len(sentence))
        
        words = []
        for i in range(1, len(lim)):
            words.append(sentence[lim[i-1]:lim[i]].strip())
        return words

    def _get_base_feat(self, word):
        """Get the base features of a word, for the CRF model.
        
        Args:
            word (str): Word to be processed.
            
        Returns:
            Dictionary with the base features of the word.
        """
        d = {
            'word': word.lower(),
            'is_title': word.istitle(),
            'is_upper': word.isupper(),
            'num_digits': str(sum(c.isdigit() for c in word)),
        }
        return d

    def _add_base_feat(self, features, sentence, index, prefix):
        """Updates a dictionary of features with the features of a word.
        
        Args:
            features (dict): Dictionary with the features already processed.
            sentence (list): List of words in the sentence.
            index (int): Index of the current word in the sentence. 
            prefix (str): Prefix to be added to the name of the features of the current word.

        """
        if index >= 0 and index < len(sentence):
            word_feat = self._get_base_feat(sentence[index])
            for feat in word_feat.keys():
                features[prefix + feat] = word_feat[feat]

    def _get_features(self, sentence):
        """Get the features of a sentence, for the CRF model.
        
        Args:
            sentence (list): List of words in the sentence.
            
        Returns:
            List of dictionaries with the features of each word.
        """
        sent_features = []
        
        for i in range(len(sentence)):
            word_feat = {
                'bias': 1.0,
                'text_position': i/len(sentence),
            }
            
            self._add_base_feat(word_feat, sentence, i-4, '-4:')
            self._add_base_feat(word_feat, sentence, i-3, '-3:')
            self._add_base_feat(word_feat, sentence, i-2, '-2:')
            self._add_base_feat(word_feat, sentence, i-1, '-1:')
            
            self._add_base_feat(word_feat, sentence, i, '')
            
            self._add_base_feat(word_feat, sentence, i+1, '+1:')
            self._add_base_feat(word_feat, sentence, i+2, '+2:')
            self._add_base_feat(word_feat, sentence, i+3, '+3:')
            self._add_base_feat(word_feat, sentence, i+4, '+4:')
                
            sent_features.append(word_feat)
        
        return sent_features

    def _extract_acts(self, text, prediction):
        """Extract and join words predicted to be part of an act.

        Args:
            text (list): List of words in the text of a DODF.
            prediction (list): Predictions made for each word in the text.
            
        Returns:
            List of acts in the text.
        """

        acts = []
        limits = self._limits(text)

        limits.append(len(text))
        prediction.append('O')

        act_start = -1
        for i in range(1, len(prediction)): # we start at 1 because sometimes the first word is tagged incorrectly
            if prediction[i][0] == 'B': #B-act
                act_start = limits[i]
            elif prediction[i][0] == 'E' and act_start != -1: #E-act
                act_end = limits[i+1]

                act = text[act_start:act_end].strip()
                if act.count('.') <= len(act)/3:
                    acts.append(act)
                act_start = -1
            
        if act_start != -1:
            acts.append(text[act_start:].strip())
            
        return acts
